from ..metamodel.compiler import Pass
from ..metamodel.typer import Checker, InferTypes

from ..metamodel.rewrite import (
    DNFUnionSplitter, FormatOutputs, HandleAggregationsAndRanks
)
from .rewrite import (
    AlgorithmPass, AnnotateConstraints, CDC, ConstantsToVars, DeduplicateVars,
    ExtractCommon, EliminateData, ExtractKeys, Flatten, FlattenScript,
    FunctionAnnotations, PeriodMath, QuantifyVars, Splinter, SplitMultiCheckRequires,
    UnifyDefinitions,
)

def lqp_passes() -> list[Pass]:
    return [
        SplitMultiCheckRequires(),
        FunctionAnnotations(),
        AnnotateConstraints(),
        Checker(),
        CDC(), # specialize to physical relations before extracting nested and typing
        InferTypes(),
        DNFUnionSplitter(), # Handle unions that require DNF decomposition
        ExtractKeys(), # Create a logical for each valid combinations of keys
        FormatOutputs(),
        ExtractCommon(), # Extracts tasks that will become common after Flatten into their own definition
        Flatten(), # Move nested tasks to the top level, and various related things touched along the way
        FlattenScript(), # Additional flattening specific to scripts
        HandleAggregationsAndRanks(), # Handle aggregation and rank dependencies
        Splinter(), # Splits multi-headed rules into multiple rules
        QuantifyVars(), # Adds missing existentials
        EliminateData(),  # Turns Data nodes into ordinary relations.
        DeduplicateVars(),  # Deduplicates vars in Updates and Outputs.
        ConstantsToVars(),  # Turns constants in Updates and Outputs into vars.
        AlgorithmPass(),
        PeriodMath(),  # Rewrite date period uses.
        UnifyDefinitions(), # Unify relations with multiple definitions.
    ]



from relationalai.semantics.lqp import ir as lqp
from relationalai.semantics.metamodel import ir, builtins
from relationalai.semantics.metamodel.helpers import sanitize
from relationalai.semantics.metamodel.util import FrozenOrderedSet

from dataclasses import dataclass
from hashlib import sha256
from typing import Tuple

class UniqueNames:
    def __init__(self):
        # Track count of seen names
        self.seen = dict[str, int]()
        # Maps id to unique name
        self.id_to_name = dict[int,str]()

    def get_name(self, name: str) -> str:
        # Names will eventually be sanitized, which could cause collisions, so we
        # do the sanitization here.
        name = '_' if name == '_' else sanitize(name)
        if name not in self.seen:
            self.seen[name] = 1
            return f"{name}"

        self.seen[name] += 1
        id = self.seen[name]
        # If the original name has a suffix we can get collisions with generated names,
        # so test the new name.
        while f"{name}_{id}" in self.seen:
            id += 1
            self.seen[name] = id
        new_name = f"{name}_{id}"
        self.seen[new_name] = 1
        return new_name

    # Get a unique name for the given id. If the id is already in the map, return the
    # existing name. Otherwise, generate a new name using the suggested_name and
    # store it in the map.
    def get_name_by_id(self, id: int, suggested_name:str) -> str:
        if id in self.id_to_name:
            return self.id_to_name[id]

        name = self.get_name(suggested_name)
        self.id_to_name[id] = name
        return name

@dataclass(frozen=True)
class ExportDescriptor:
    relation_id: lqp.RelationId
    column_name: str
    column_number: int
    column_type: lqp.Type
    export_fqn: str

# Type alias for export_info structure: list of (export_fqn, filename, column_types)
ExportInfo = list[tuple[str, str, list]]

def extract_export_fqn(annotations: list[ir.Annotation] | FrozenOrderedSet[ir.Annotation]) -> str:
    """
    Extract the export FQN from annotations.

    Returns the export FQN string if an export annotation is found, "unnamed_table" otherwise.
    """
    for annotation in annotations:
        if annotation.relation == builtins.export:
            if annotation.args:
                export_fqn_literal = annotation.args[0]
                if isinstance(export_fqn_literal, ir.Literal) and isinstance(export_fqn_literal.value, str):
                    return export_fqn_literal.value
    return "unnamed_table"

class TranslationCtx:
    def __init__(self, def_names: UniqueNames = UniqueNames()):
        # TODO: comment these fields
        self.def_names = def_names
        self.var_names = UniqueNames()
        self.output_names = UniqueNames()
        # A counter for break rules generated during translation of while loops
        self.break_rule_counter = 0
        # Map relation IDs to their original names for debugging and pretty printing.
        self.rel_id_to_orig_name = {}
        self.output_ids: list[tuple[lqp.RelationId, str]] = []
        self.export_descriptors: list[ExportDescriptor] = []

def gen_rel_id(ctx: TranslationCtx, orig_name: str, suffix: str = "") -> lqp.RelationId:
    relation_id = lqp.RelationId(id=lqp_hash(orig_name + suffix), meta=None)
    ctx.rel_id_to_orig_name[relation_id] = orig_name
    return relation_id

def gen_unique_var(ctx: TranslationCtx, name_hint: str) -> lqp.Var:
    """
    Generate a new variable with a unique name based on the provided hint.
    """
    name = ctx.var_names.get_name(name_hint)
    return lqp.Var(name=name, meta=None)

def is_constant(arg, expected_type):
    """
    Check if the argument is a constant of the expected type.
    """
    if isinstance(arg, ir.Literal):
        return is_constant(arg.value, expected_type)

    return isinstance(arg, expected_type)

def rename_vars_var(var: lqp.Var, var_map: dict[str, lqp.Var]) -> lqp.Var:
    return var_map.get(var.name, var)

def rename_vars_relterm(term: lqp.RelTerm, var_map: dict[str, lqp.Var]) -> lqp.RelTerm:
    if isinstance(term, lqp.Var):
        return rename_vars_var(term, var_map)
    else:
        return term  # Constants do not change

def rename_vars_term(term: lqp.Term, var_map: dict[str, lqp.Var]) -> lqp.Term:
    if isinstance(term, lqp.Var):
        return rename_vars_var(term, var_map)
    else:
        return term  # Constants do not change

def rename_vars_abstraction(abstraction: lqp.Abstraction, var_map: dict[str, lqp.Var]) -> lqp.Abstraction:
    new_vars = [(var_map.get(var.name, var), typ) for (var, typ) in abstraction.vars]
    new_value = rename_vars_formula(abstraction.value, var_map)
    return lqp.Abstraction(vars=new_vars, value=new_value, meta=abstraction.meta)

def rename_vars_formula(formula: lqp.Formula, var_map: dict[str, lqp.Var]) -> lqp.Formula:
    if isinstance(formula, lqp.Primitive):
        return lqp.Primitive(
            name=formula.name,
            terms=[rename_vars_relterm(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Atom):
        return lqp.Atom(
            name=formula.name,
            terms=[rename_vars_term(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.RelAtom):
        return lqp.RelAtom(
            name=formula.name,
            terms=[rename_vars_relterm(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Not):
        return lqp.Not(arg=rename_vars_formula(formula.arg, var_map), meta=formula.meta)
    elif isinstance(formula, lqp.Exists):
        return lqp.Exists(body=rename_vars_abstraction(formula.body, var_map), meta=None)
    elif isinstance(formula, lqp.Reduce):
        return lqp.Reduce(
            op=formula.op,
            body=rename_vars_abstraction(formula.body, var_map),
            terms=[rename_vars_term(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.FFI):
        return lqp.FFI(
            meta=formula.meta,
            name=formula.name,
            args=[rename_vars_abstraction(arg, var_map) for arg in formula.args],
            terms=[rename_vars_term(term, var_map) for term in formula.terms]
        )
    elif isinstance(formula, lqp.Conjunction):
        return lqp.Conjunction(
            args=[rename_vars_formula(arg, var_map) for arg in formula.args],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Disjunction):
        return lqp.Disjunction(
            args=[rename_vars_formula(arg, var_map) for arg in formula.args],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Cast):
        return lqp.Cast(
            input=rename_vars_term(formula.input, var_map),
            result=rename_vars_term(formula.result, var_map),
            meta=formula.meta,
        )
    else:
        raise NotImplementedError(f"Unknown formula type: {type(formula)}")

def lqp_hash(node: str) -> int:
    h = int.from_bytes(sha256(node.encode()).digest(), byteorder='big', signed=False)
    # Ensure it's within the 128-bit range
    return h % (2**128)

def output_names(aliases: FrozenOrderedSet[Tuple[str, ir.Value]]) -> list[str]:
    return [v[0] for v in aliases]



from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor

from typing import cast
import pandas as pd
import hashlib

# Creates intermediary relations for all Data nodes and replaces said Data nodes
# with a Lookup into these created relations. Reuse duplicate created relations.
class EliminateData(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.DataRewriter()
        return r.walk(model)

    # Does the actual work.
    class DataRewriter(visitor.Rewriter):
        new_relations: list[ir.Relation]
        new_updates: list[ir.Logical]
        # Counter for naming new relations.
        # It must be that new_count == len new_updates == len new_relations.
        new_count: int
        # Cache for Data nodes to avoid creating duplicate intermediary relations
        data_cache: dict[str, ir.Relation]

        def __init__(self):
            self.new_relations = []
            self.new_updates = []
            self.new_count = 0
            self.data_cache = {}
            super().__init__()

        # Create a cache key for a Data node based on its structure and content
        def _data_cache_key(self, node: ir.Data) -> str:
            values = pd.util.hash_pandas_object(node.data).values
            return hashlib.sha256(bytes(values)).hexdigest()

        def _intermediary_relation(self, node: ir.Data) -> ir.Relation:
            cache_key = self._data_cache_key(node)
            if cache_key in self.data_cache:
                return self.data_cache[cache_key]
            self.new_count += 1
            intermediary_name = f"formerly_Data_{self.new_count}"

            intermediary_relation = f.relation(
                intermediary_name,
                [f.field(v.name, v.type) for v in node.vars]
            )
            self.new_relations.append(intermediary_relation)

            intermediary_update = f.logical([
                # For each row (union), equate values and their variable (logical).
                f.union(
                    [
                        f.logical(
                            [
                                f.lookup(rel_builtins.eq, [f.literal(val, var.type), var])
                                for (val, var) in zip(row, node.vars)
                            ],
                        )
                        for row in node
                    ],
                    hoisted = node.vars,
                ),
                # And pop it back into the relation.
                f.update(intermediary_relation, node.vars, ir.Effect.derive),
            ])
            self.new_updates.append(intermediary_update)

            # Cache the result for reuse
            self.data_cache[cache_key] = intermediary_relation

            return intermediary_relation

        # Create a new intermediary relation representing the Data (and pop it in
        # new_updates/new_relations) and replace this Data with a Lookup of said
        # intermediary.
        def handle_data(self, node: ir.Data, parent: ir.Node) -> ir.Lookup:
            intermediary_relation = self._intermediary_relation(node)
            replacement_lookup = f.lookup(intermediary_relation, node.vars)

            return replacement_lookup

        # Walks the model for the handle_data work then updates the model with
        # the new state.
        def handle_model(self, model: ir.Model, parent: None):
            walked_model = super().handle_model(model, parent)
            assert len(self.new_relations) == len(self.new_updates) and self.new_count == len(self.new_relations)

            # This is okay because its LQP.
            assert isinstance(walked_model.root, ir.Logical)
            root_logical = cast(ir.Logical, walked_model.root)

            # We may need to add the new intermediaries from handle_data to the model.
            if self.new_count  == 0:
                return model
            else:
                return ir.Model(
                    walked_model.engines,
                    walked_model.relations | self.new_relations,
                    walked_model.types,
                    ir.Logical(
                        root_logical.engine,
                        root_logical.hoisted,
                        root_logical.body + tuple(self.new_updates),
                        root_logical.annotations,
                    ),
                    walked_model.annotations,
                )
